/*
 * style.css: default style sheet
 *
 * Copyright (c) 2003 Per Cederberg. Permission is granted to copy this
 * document verbatim in any medium, provided that this copyright notice
 * is left intact.
 */



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Default Styles
 *
 */

/* Normal document settings */
body {
        margin-top: 2cm;
        margin-bottom: 2cm;
        margin-left: 2.5cm;
        margin-right: 2.5cm;
        background: white;
        color: #331100;
        font-family: helvetica;
}

/* Header formats */
h1 {
        font-size: 160%;
        color: rgb(180, 180, 180);
}
h2 {
        font-size: 120%;
        margin-top: 2em;
        margin-bottom: 0.5em;
        padding: 3px;
        padding-left: 7px;
        background: rgb(230, 230, 230);
        color: rgb(0, 49, 99);
}
h3 {
        font-size: 100%;
        margin-top: 1.5em;
        margin-bottom: 0.5em;
        padding: 3px;
        padding-left: 7px;
        background: rgb(185, 185, 185);
        color: white;
}

/* Separator style */
hr {
        margin-top: 2em;
        width: 100%;
        height: 1px;
        color: black;
}

/* Link styles */
a {
        font-weight: bold;
}
a:link {
        color: rgb(100, 150, 200);
}
a:visited {
        color: rgb(0, 49, 99);
}
a:active {
        background: rgb(240, 240, 255);
        color: rgb(50, 100, 150);
}
a:hover {
        background: rgb(230, 230, 255);
}
a img {
        border-style: none;
}

/* Image style */
img {
        vertical-align: middle;
}

/* Code example style */
pre {
        font-size: 100%;
        font-weight: bold;
        color: rgb(0, 49, 99);
}
code {
        font-size: 100%;
        font-weight: bold;
        color: rgb(0, 49, 99);
}

/* Script styles */
sub {
        color: rgb(50, 100, 150);
}
sup {
        color: rgb(50, 100, 150);
}

/* Table style */
table {
        width: auto;
        border: none;
}
th {
        padding: 5px;
        padding-left: 8px;
        padding-right: 8px;
        vertical-align: top;
        text-align: left;
        border: none;
}
td {
        padding: 5px;
        padding-left: 8px;
        padding-right: 8px;
        vertical-align: top;
        border: none;
}

/* List styles */
ul,
ol,
dl {
        margin-top: 0.5em;
        margin-bottom: 0.5em;
}
dt {
        margin-top: 0.7em;
        margin-bottom: 0em;
        margin-left: 30px;
}
dd {
        font-size: smaller;
        margin-left: 30px;
        margin-top: 0.4em;
        margin-bottom: 1em;
}
li {
        margin-bottom: 1em;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  Special Styles
 *
 */

/* Menu styles */
p.menu {
        margin-top: 0em;
        margin-bottom: 2em;
        font-size: smaller;
}
table.menu {
        margin-bottom: 1em;
        border-collapse: collapse;
        font-size: smaller;
}
table.menu td.active {
        border: 1px solid black;
        border-bottom-color: white;
        padding-left: 0.7em;
        padding-right: 0.7em;
        background: white;
}
table.menu td.inactive {
        border: 1px solid black;
        padding-left: 0.7em;
        padding-right: 0.7em;
        background: rgb(230, 230, 230);
}
table.menu td.space {
        width: 0.3em;
        border-bottom: 1px solid black;
        padding: 0px;
}
table.menu td.filler {
        width: 100%;
        border-bottom: 1px solid black;
        padding: 0px;
}

/* Compact table style */
table.compact {
        border: none;
}
table.compact th {
        padding: 0px;
        padding-left: 5px;
        padding-right: 5px;
}
table.compact td {
        padding: 0px;
        padding-left: 5px;
        padding-right: 5px;
}

/* Bordered table style */
table.border {
        border-style: solid;
        border-width: 1px;
        border-color: black;
        border-collapse: collapse;
}
table.border th {
        border-style: solid;
        border-width: 1px;
        border-color: darkgray;
}
table.border td {
        border-style: solid;
        border-width: 1px;
        border-color: darkgray;
}

/* Paragraph styles */
.first {
        margin-top: 0px;
}

/* Semantic markers */
.header {
        font-size: smaller;
}
.footer {
        font-size: smaller;
}
.leader {
        font-size: smaller;
        font-weight: bold;
        color: rgb(50, 100, 150);
}
.figure {
        background: rgb(255,255,245);
        margin: 6px;
        padding: 0.5cm;
        padding-top: 0.1cm;
        padding-bottom: 0.3cm;
        border-style: solid;
        border-width: 1px;
}
.example {
        padding-left: 1.5em;
}
.illustration {
        text-align: center;
}
.source {
        color: rgb(0, 49, 99);
}
.important {
        padding: 3px;
        background: #ffcccc;
        color: red;
        font-weight: bolder;
        font-style: italic;
}
.important a,
.important a:link,
.important a:visited,
.important a:active {
        background: rgb(255, 204, 204);
        color: red;
}
.unimportant {
        color: rgb(170, 170, 170);
}
.correct {
        color: green;
}
.incorrect {
        color: red;
}
.draft {
        background: rgb(255, 204, 204);
        color: red;
        padding-left: 3px;
        padding-right: 3px;
}
