/*
 * RegexpConstants.java
 *
 * THIS FILE HAS BEEN GENERATED AUTOMATICALLY. DO NOT EDIT!
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * LICENSE.txt file for more details.
 *
 * Copyright (c) 2003-2015 Per Cederberg. All rights reserved.
 */

package net.percederberg.grammatica.test;

/**
 * An interface with constants for the parser and tokenizer.
 *
 * @author   Per Cederberg
 * @version  1.0
 */
interface RegexpConstants {

    /**
     * A token identity constant.
     */
    public static final int LEFT_PAREN = 1001;

    /**
     * A token identity constant.
     */
    public static final int RIGHT_PAREN = 1002;

    /**
     * A token identity constant.
     */
    public static final int LEFT_BRACKET = 1003;

    /**
     * A token identity constant.
     */
    public static final int RIGHT_BRACKET = 1004;

    /**
     * A token identity constant.
     */
    public static final int LEFT_BRACE = 1005;

    /**
     * A token identity constant.
     */
    public static final int RIGHT_BRACE = 1006;

    /**
     * A token identity constant.
     */
    public static final int QUESTION = 1007;

    /**
     * A token identity constant.
     */
    public static final int ASTERISK = 1008;

    /**
     * A token identity constant.
     */
    public static final int PLUS = 1009;

    /**
     * A token identity constant.
     */
    public static final int VERTICAL_BAR = 1010;

    /**
     * A token identity constant.
     */
    public static final int DOT = 1011;

    /**
     * A token identity constant.
     */
    public static final int COMMA = 1012;

    /**
     * A token identity constant.
     */
    public static final int NUMBER = 1013;

    /**
     * A token identity constant.
     */
    public static final int CHAR = 1014;

    /**
     * A production node identity constant.
     */
    public static final int EXPR = 2001;

    /**
     * A production node identity constant.
     */
    public static final int TERM = 2002;

    /**
     * A production node identity constant.
     */
    public static final int FACT = 2003;

    /**
     * A production node identity constant.
     */
    public static final int ATOM = 2004;

    /**
     * A production node identity constant.
     */
    public static final int ATOM_MODIFIER = 2005;

    /**
     * A production node identity constant.
     */
    public static final int CHARACTER_SET = 2006;

    /**
     * A production node identity constant.
     */
    public static final int CHARACTER = 2007;
}
