/*
 * RegexpConstants.cs
 *
 * THIS FILE HAS BEEN GENERATED AUTOMATICALLY. DO NOT EDIT!
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * LICENSE.txt file for more details.
 *
 * Copyright (c) 2003-2015 Per Cederberg. All rights reserved.
 */

namespace PerCederberg.Grammatica.Test {

    /**
     * <remarks>An enumeration with token and production node
     * constants.</remarks>
     */
    internal enum RegexpConstants {
        LEFT_PAREN = 1001,
        RIGHT_PAREN = 1002,
        LEFT_BRACKET = 1003,
        RIGHT_BRACKET = 1004,
        LEFT_BRACE = 1005,
        RIGHT_BRACE = 1006,
        QUESTION = 1007,
        ASTERISK = 1008,
        PLUS = 1009,
        VERTICAL_BAR = 1010,
        DOT = 1011,
        COMMA = 1012,
        NUMBER = 1013,
        CHAR = 1014,
        EXPR = 2001,
        TERM = 2002,
        FACT = 2003,
        ATOM = 2004,
        ATOM_MODIFIER = 2005,
        CHARACTER_SET = 2006,
        CHARACTER = 2007
    }
}
