/*
 * AssemblyInfo.cs
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * LICENSE.txt file for more details.
 *
 * Copyright (c) 2003-2015 Per Cederberg. All rights reserved.
 */

using System.Reflection;
using System.Runtime.CompilerServices;

[assembly: AssemblyTitle("PerCederberg.Grammatica")]
[assembly: AssemblyDescription("Grammatica Run-Time Library")]
[assembly: AssemblyConfiguration("")]
[assembly: AssemblyCompany("http://grammatica.percederberg.net/")]
[assembly: AssemblyProduct("Grammatica")]
[assembly: AssemblyCopyright("2003-2015 Per Cederberg")]
[assembly: AssemblyTrademark("")]
[assembly: AssemblyCulture("")]
[assembly: AssemblyVersion("1.6.0.*")]
[assembly: AssemblyDelaySign(false)]
[assembly: AssemblyKeyFile("")]
[assembly: AssemblyKeyName("")]
